@@ -10,6 +10,7 @@ TESTDIR=testdir
 
 PSQL=$PGBIN/psql
 PGPROTO=$PGPOOL_INSTALL_DIR/bin/pgproto
+PCP_INVALIDATE_QUERY_CACHE=$PGPOOL_INSTALL_DIR/bin/pcp_invalidate_query_cache
 
 # remove error/notice details (message and so on) from
 # ErrorResponse or NoticeResponse messages.
@@ -91,6 +92,10 @@ SELECT '2022-07-05 10:00:00'::TIMETZ;
 SELECT '2022-07-05 10:00:00'::TIMETZ;
 SELECT to_timestamp(0);
 SELECT to_timestamp(0);
+/*FORCE QUERY CACHE*/SELECT now();
+/*FORCE QUERY CACHE*/SELECT now();
+/*NO QUERY CACHE*/SELECT 1;
+/*NO QUERY CACHE*/SELECT 1;
 EOF
 
 	success=true
@@ -110,6 +115,8 @@ EOF
 	grep "fetched from cache" log/pgpool.log | grep 'TIMESTAMPTZ;' > /dev/null && success=false
 	grep "fetched from cache" log/pgpool.log | grep 'TIMETZ;' > /dev/null && success=false
 	grep "fetched from cache" log/pgpool.log | grep 'to_timestamp' > /dev/null && success=false
+	grep "fetched from cache" log/pgpool.log | grep 'FORCE QUERY CACHE' > /dev/null || success=false
+	grep "fetched from cache" log/pgpool.log | grep 'NO QUERY CACHE' > /dev/null && success=false
 
 	if [ $success = false ];then
 		./shutdownall
@@ -451,6 +458,19 @@ ALTER ROLE foo WITH CONNECTION LIMIT 10;
 SELECT 10;
 EOF
 
+		$PSQL -a test >> result 2>&1 <<EOF
+--
+-- PGPOOL SET CACHE DELETE test cases.
+--
+-- force to create cache
+/*FORCE QUERY CACHE*/SELECT 1;
+-- make sure the cache was created
+/*FORCE QUERY CACHE*/SELECT 1;
+-- delete the cache
+PGPOOL SET CACHE DELETE '/*FORCE QUERY CACHE*/SELECT 1;';
+-- make sure the cache was deleted
+/*FORCE QUERY CACHE*/SELECT 1;
+EOF
 	./shutdownall
 
 	cd ..
@@ -465,6 +485,94 @@ EOF
 	    exit 1
 	fi
 	rm $log
+
+	cd $TESTDIR
+	./startall
+	wait_for_pgpool_startup
+
+	# test for pcp_invalidate_query_cache
+	res1=`$PSQL -t -c "/*FORCE QUERY CACHE*/SELECT current_timestamp" test`
+	res2=`$PSQL -t -c "/*FORCE QUERY CACHE*/SELECT current_timestamp" test`
+	# make sure query cache created
+	if [ "$res1" != "$res2" ];then
+	    echo "query cache was not created in pcp_invalidate_query_cache test"
+	    ./shutdownall
+	    exit 1
+	fi
+	# remove query cache
+	$PCP_INVALIDATE_QUERY_CACHE -p $PCP_PORT
+	if [ $? != 0 ];then
+	    echo "pcp_invalidate_query_cache failed"
+	    ./shutdownall
+	    exit 1
+	fi
+	# make sure query cache has gone
+	$PSQL -t -c "SELECT 1" test	# this query processes query cache invalidation request
+	res1=`$PSQL -t -c "/*FORCE QUERY CACHE*/SELECT current_timestamp" test`
+	if [ "$res1" = "$res2" ];then
+	    echo "query cache was not invalidated"
+	    ./shutdownall
+	    exit 1
+	fi
+	./shutdownall
+
+	cd ..
+done
+
+#
+# Test for extended query protocol coner cases in streaming replication mode.
+# These tests are basically for a sequence of extended queries:
+# 1. execute a SELECT and create query cache entry.
+# 2. sync.
+# 3. execute another a SELECT.
+# 4. execute bind and execute to use the query cache created at #1.
+# 5. sync.
+
+rm -fr $TESTDIR
+mkdir $TESTDIR
+cd $TESTDIR
+
+# create test environment
+echo -n "creating test environment..."
+$PGPOOL_SETUP -m s -n 2 || exit 1
+echo "done."
+
+echo "memory_cache_enabled = on" >> etc/pgpool.conf
+cd ..
+
+for i in 1 2 3 4 4
+do
+    #
+    # case 1: failed with kind mismatch error at #5.
+    # "packet kind of backend 0 ['T'] does not match with main/majority nodes packet kind ['Z']"
+    #
+    # case 2: step #4 includes error (hung).
+    #
+    # case 3: step #4 includes PortalSuspended (hung).
+    #
+    # case 4: various cases including portal suspended
+    # Note that case4 is executed twice to make sure that
+    # the test works for either query cache exists or does not exist
+    cd $TESTDIR
+    ./startall
+    wait_for_pgpool_startup
+    timeout 1 $PGPROTO -d test -f ../query_cache_bug$i.data |& del_details_from_error > result
+    if [ $? != 0 ];then
+	# timeout happened or pgproto returned non 0 status
+	echo "test failed in test case #2 (timeout)"
+	err=true
+	./shutdownall
+	exit 1
+    fi
+    ./shutdownall
+    cd ..
+    diff -c expected.$i $TESTDIR/result > $log
+    if [ $? != 0 ];then
+	echo "test failed in test case $i"
+	cat $log
+	rm $log
+	exit 1
+    fi
 done
 
 exit 0
